from __future__ import annotations

from typing import TYPE_CHECKING
import json

from youtube_unofficial.client import NoFeedbackToken, YouTubeClient
from youtube_unofficial.constants import WATCH_HISTORY_URL
import pytest

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture
    from requests_mock import Mocker


def test_clear_watch_history_no_feedback_token(mocker: MockerFixture, requests_mock: Mocker,
                                               client: YouTubeClient) -> None:
    mocker.patch('youtube_unofficial.client.Soup')
    mocker.patch('youtube_unofficial.client.find_ytcfg', return_value={})
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value={
                     'contents': {
                         'twoColumnBrowseResultsRenderer': {
                             'secondaryContents': {
                                 'browseFeedActionsRenderer': {}
                             }
                         }
                     }
                 })
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    requests_mock.post(
        'https://www.youtube.com/youtubei/v1/feedback',
        json={'status': 'STATUS_SUCCEEDED'},
    )

    with pytest.raises(NoFeedbackToken):
        client.clear_watch_history()


def test_clear_watch_history_clear_button_disabled(mocker: MockerFixture, requests_mock: Mocker,
                                                   client: YouTubeClient, data_path: Path) -> None:
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    mocker.patch('youtube_unofficial.client.Soup')
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'USER_SESSION_ID': 'test_session_id',
                     'INNERTUBE_CONTEXT_CLIENT_VERSION': '1.0'
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'clear-watch-history/00-button-disabled.json').read_text()))
    assert client.clear_watch_history() is False


def test_clear_watch_history(mocker: MockerFixture, requests_mock: Mocker, client: YouTubeClient,
                             data_path: Path) -> None:
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    requests_mock.post('https://www.youtube.com/youtubei/v1/feedback',
                       json={'feedbackResponses': [{
                           'isProcessed': True
                       }]})
    mocker.patch('youtube_unofficial.client.Soup')
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'USER_SESSION_ID': 'test_session_id',
                     'INNERTUBE_CONTEXT_CLIENT_VERSION': '1.0'
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads((data_path / 'clear-watch-history/00.json').read_text()))
    assert client.clear_watch_history() is True



"""Entry point for the CLI."""
from __future__ import annotations

import click

from .commands import (
    clear_watch_history,
    clear_watch_later,
    print_history,
    print_playlist,
    print_watch_later,
    remove_history_entries,
    remove_video_id,
    remove_watch_later_video_id,
    toggle_watch_history,
)


@click.group(context_settings={'help_option_names': ('-h', '--help')})
def main() -> None:
    """Unofficial YouTube CLI."""


main.add_command(clear_watch_history)
main.add_command(clear_watch_later)
main.add_command(print_history)
main.add_command(print_playlist)
main.add_command(print_watch_later)
main.add_command(remove_history_entries)
main.add_command(remove_video_id)
main.add_command(remove_watch_later_video_id)
main.add_command(toggle_watch_history)



from __future__ import annotations

from typing import TYPE_CHECKING
import json

from youtube_unofficial.constants import WATCH_HISTORY_URL

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture
    from requests_mock import Mocker
    from youtube_unofficial.client import YouTubeClient


def test_toggle_history(mocker: MockerFixture, client: YouTubeClient, requests_mock: Mocker,
                        data_path: Path) -> None:
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    requests_mock.post('https://www.youtube.com/youtubei/v1/feedback',
                       json={'feedbackResponses': [{
                           'isProcessed': True
                       }]})
    mock_session = mocker.MagicMock()
    mocker.patch('yt_dlp_utils.setup_session', return_value=mock_session)
    mocker.patch('youtube_unofficial.client.Soup')
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'USER_SESSION_ID': 'test_session_id',
                     'INNERTUBE_CONTEXT_CLIENT_VERSION': '1.0'
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads((data_path / 'toggle-history-00.json').read_text()))
    result = client.toggle_watch_history()
    assert result is True



"""Configuration for Pytest."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, NoReturn, cast
import os

from click.testing import CliRunner
from youtube_unofficial.client import YouTubeClient
import pytest

if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from pytest_mock import MockerFixture

if os.getenv('_PYTEST_RAISE', '0') != '0':  # pragma no cover

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call: pytest.CallInfo[None]) -> NoReturn:
        assert call.excinfo is not None
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo: pytest.ExceptionInfo[BaseException]) -> NoReturn:
        raise excinfo.value


@pytest.fixture
def runner() -> CliRunner:
    return CliRunner()


@pytest.fixture
def mock_cookie_jar(mocker: MockerFixture) -> MagicMock:
    mock_jar = mocker.MagicMock()
    mock_jar.get.return_value = 'test_cookie'
    return cast('MagicMock', mock_jar)


@pytest.fixture
def mock_extract_cookies_from_browser(mocker: MockerFixture) -> None:
    mock_jar = mocker.MagicMock()
    mocker.patch('yt_dlp_utils.lib.extract_cookies_from_browser', return_value=mock_jar)


@pytest.fixture
def client(mock_cookie_jar: MagicMock, mock_extract_cookies_from_browser: None) -> YouTubeClient:
    client = YouTubeClient(browser='firefox', profile='default')
    client.session.cookies = mock_cookie_jar
    return client


@pytest.fixture
def data_path() -> Path:
    return Path(__file__).parent / 'client/data'



"""Typed dictionaries for YouTube playlist data."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypedDict

if TYPE_CHECKING:
    from collections.abc import Sequence

__all__ = (
    'HasKeyText',
    'Icon',
    'NavigationEndpoint',
    'NextContinuation',
    'PlaylistInfo',
    'PlaylistVideoIDsEntry',
    'PlaylistVideoListRenderer',
    'PlaylistVideoListRendererContinuations',
    'PlaylistVideoRenderer',
    'PlaylistVideoRendererMenu',
    'PlaylistVideoRendererMenuRenderer',
    'PlaylistVideoRendererMenuRendererItem',
    'PlaylistVideoRendererMenuRendererItemMenuServiceItemRenderer',
    'RunsOrSimpleText',
    'RunsOrText',
    'WatchEndpoint',
)


class NextContinuation(TypedDict):
    """Continuation data."""
    clickTrackingParams: str
    """Click tracking parameters."""
    continuation: str
    """Continuation token."""


class WatchEndpoint(TypedDict):
    """Watch endpoint data."""
    videoId: str
    """Video ID."""


class NavigationEndpoint(TypedDict):
    """Navigation endpoint data."""
    watchEndpoint: WatchEndpoint
    """Watch endpoint data."""


class HasKeyText(TypedDict):
    """Inside :py:class:`RunsOrSimpleText` and :py:class:`RunsOrText`."""
    text: str
    """Text content."""


class RunsOrSimpleText(TypedDict, total=False):
    """Inside :py:class:`PlaylistVideoRenderer`."""
    runs: Sequence[HasKeyText]
    """Text runs."""
    simpleText: str
    """Text content."""


class RunsOrText(TypedDict, total=False):
    """Inside :py:class:`PlaylistVideoRenderer`."""
    runs: Sequence[HasKeyText]
    """Text runs."""
    text: str
    """Text content."""


class Icon(TypedDict):
    """Icon data."""
    iconType: str
    """Icon type."""


class PlaylistVideoRendererMenuRendererItemMenuServiceItemRenderer(TypedDict):
    """Menu renderer."""
    icon: Icon
    """Icon data."""
    serviceEndpoint: dict[str, Any]
    """Service endpoint data."""


class PlaylistVideoRendererMenuRendererItem(TypedDict):
    """Menu renderer item."""
    menuServiceItemRenderer: PlaylistVideoRendererMenuRendererItemMenuServiceItemRenderer
    """Menu service item renderer."""


class PlaylistVideoRendererMenuRenderer(TypedDict):
    """Menu renderer."""
    items: Sequence[PlaylistVideoRendererMenuRendererItem]
    """Menu items."""


class PlaylistVideoRendererMenu(TypedDict):
    """Menu data."""
    menuRenderer: PlaylistVideoRendererMenuRenderer
    """Menu renderer data."""


class PlaylistVideoRenderer(TypedDict, total=False):
    """Playlist video renderer dictionary."""
    menu: PlaylistVideoRendererMenu
    """Menu data."""
    navigationEndpoint: NavigationEndpoint
    """Navigation endpoint data."""
    shortBylineText: RunsOrText
    """Short byline text."""
    title: RunsOrSimpleText
    """Video title."""
    videoId: str
    """Video ID."""


class PlaylistInfo(TypedDict, total=False):
    """Playlist information."""
    continuationItemRenderer: dict[str, Any]
    """Continuation data."""
    playlistVideoRenderer: PlaylistVideoRenderer
    """Inner renderer data."""


class PlaylistVideoListRendererContinuations(TypedDict):
    """Continuation data."""
    nextContinuationData: NextContinuation
    """Continuation data."""


class PlaylistVideoListRenderer(TypedDict):
    """Playlist video list renderer dictionary."""
    contents: Sequence[PlaylistInfo]
    """Inner contents."""
    continuations: Sequence[PlaylistVideoListRendererContinuations]
    """Continuation data."""


class PlaylistVideoIDsEntry(TypedDict):
    """Playlist information."""
    owner: str | None
    """Playlist owner."""
    title: str | None
    """Playlist title."""
    video_id: str
    """Video ID."""
    watch_url: str
    """Watch URL."""



from __future__ import annotations

from typing import TYPE_CHECKING

from bs4 import BeautifulSoup
from youtube_unofficial.utils import (
    context_client_body,
    extract_keys,
    extract_script_content,
    find_ytcfg,
    get_text_runs,
    initial_data,
    ytcfg_headers,
)

if TYPE_CHECKING:
    from pytest_mock import MockerFixture
    from youtube_unofficial.typing.history import DescriptionSnippet
    from youtube_unofficial.typing.ytcfg import YtcfgDict


def test_extract_keys() -> None:
    keys = ['a', 'b']
    obj = {'a': 1, 'b': 2, 'c': 3}
    result = extract_keys(keys, obj)
    assert result == {'a': 1, 'b': 2}


def test_get_text_runs() -> None:
    desc: DescriptionSnippet = {'runs': [{'text': 'Hello'}, {'text': 'World'}]}
    result = get_text_runs(desc)
    assert result == 'HelloWorld'


def test_context_client_body() -> None:
    ytcfg: YtcfgDict = {'INNERTUBE_CONTEXT_CLIENT_VERSION': '1.20230101'}
    result = context_client_body(ytcfg)
    assert result == {'clientName': 'WEB', 'clientVersion': '1.20230101'}


def test_initial_data(mocker: MockerFixture) -> None:
    mock_content = mocker.MagicMock(spec=BeautifulSoup)
    mock_script = 'var ytInitialData = {"key": "value"};'
    mocker.patch('youtube_unofficial.utils.extract_script_content', return_value=[mock_script])
    result = initial_data(mock_content)
    assert result == {'key': 'value'}


def test_find_ytcfg(mocker: MockerFixture) -> None:
    mock_soup = mocker.MagicMock(spec=BeautifulSoup)
    mock_script = '<script>ytcfg.set({"INNERTUBE_CONTEXT_CLIENT_VERSION": "1.20230101"});</script>'
    mocker.patch('youtube_unofficial.utils.extract_script_content', return_value=[mock_script])
    result = find_ytcfg(mock_soup)
    assert result == {'INNERTUBE_CONTEXT_CLIENT_VERSION': '1.20230101'}


def test_ytcfg_headers() -> None:
    ytcfg: YtcfgDict = {'USER_SESSION_ID': '12345'}
    result = ytcfg_headers(ytcfg)
    assert result == {
        'x-goog-authuser': '0',
        'x-goog-page-id': '12345',
        'x-origin': 'https://www.youtube.com',
    }


def test_extract_script_content(mocker: MockerFixture) -> None:
    mock_soup = mocker.MagicMock(spec=BeautifulSoup)
    mock_script_tags = [
        mocker.MagicMock(children=['var ytInitialData = {"key": "value"};']),
        mocker.MagicMock(children=['<script>console.log("test");</script>']),
    ]
    mock_soup.select.return_value = mock_script_tags

    result = list(extract_script_content(mock_soup))

    assert result == [
        'var ytInitialData = {"key": "value"};',
        '<script>console.log("test");</script>',
    ]
    mock_soup.select.assert_called_once_with('script')



from __future__ import annotations

from typing import TYPE_CHECKING
import json

from youtube_unofficial.constants import WATCH_HISTORY_URL

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture
    from requests_mock import Mocker
    from youtube_unofficial.client import YouTubeClient


def test_response_no_feedback_responses_key(mocker: MockerFixture, requests_mock: Mocker,
                                            client: YouTubeClient, data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'USER_SESSION_ID': 'test_session_id',
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'INNERTUBE_CONTEXT_CLIENT_VERSION': '1.0',
                     'SESSION_INDEX': 0,
                     'VISITOR_DATA': 'test_visitor_data',
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads((data_path / 'response-no-feedback-00.json').read_text()))
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    requests_mock.post('https://www.youtube.com/youtubei/v1/feedback', json={})
    assert client.remove_video_ids_from_history(['test_video']) is False



from __future__ import annotations

from typing import TYPE_CHECKING
import json

from youtube_unofficial.constants import WATCH_HISTORY_URL

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture
    from requests_mock import Mocker
    from youtube_unofficial.client import YouTubeClient


def test_get_history_video_ids(mocker: MockerFixture, requests_mock: Mocker, client: YouTubeClient,
                               data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'VISITOR_DATA': 'test_visitor_data',
                     'USER_SESSION_ID': 'test_session_id',
                     'SESSION_INDEX': 0,
                 })
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads((data_path / 'get-history-video-ids/00.json').read_text()))
    result = list(client.get_history_video_ids(return_dict=True))
    assert result == [{
        'video_id': 'test_video_id',
        'title': 'Test Title',
        'owner_text': 'Test Owner',
        'short_byline_text': 'Test Channel',
        'length': '5:00',
        'length_accessible': '5 minutes',
        'verified': True,
        'short_view_count_text': '10 views',
        'moving_thumbnails': [{
            'height': 100,
            'url': 'https://example.com/thumbnail.jpg',
            'width': 100
        }],
        'video_thumbnails': [{
            'height': 100,
            'url': 'https://example.com/thumbnail.jpg',
            'width': 100
        }, {
            'height': 320,
            'width': 180,
            'url': 'test_thumbnail_url',
        }],
        'random': 1,
        'watch_url': 'https://www.youtube.com/watch?v=test_video_id',
    }, {
        'video_id': 'test_video_id',
        'title': 'Test Title',
        'owner_text': 'Test Owner',
        'short_byline_text': 'Test Channel',
        'length': '5:00',
        'length_accessible': '5 minutes',
        'verified': False,
        'short_view_count_text': '10 views',
        'moving_thumbnails': [{
            'height': 100,
            'url': 'https://example.com/thumbnail.jpg',
            'width': 100
        }],
        'video_thumbnails': [{
            'height': 100,
            'url': 'https://example.com/thumbnail.jpg',
            'width': 100
        }, {
            'height': 320,
            'width': 180,
            'url': 'test_thumbnail_url',
        }],
        'random': 1,
        'watch_url': 'https://www.youtube.com/watch?v=test_video_id',
    }]


def test_get_history_video_ids_strings(mocker: MockerFixture, requests_mock: Mocker,
                                       client: YouTubeClient, data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'VISITOR_DATA': 'test_visitor_data',
                     'USER_SESSION_ID': 'test_session_id',
                     'SESSION_INDEX': 0,
                 })
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'get-history-video-ids/00-strings.json').read_text()))
    result = list(client.get_history_video_ids())
    assert result == ['test_video_id', 'test_video_id']


def test_get_history_video_ids_empty(mocker: MockerFixture, requests_mock: Mocker,
                                     client: YouTubeClient, data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'VISITOR_DATA': 'test_visitor_data',
                     'USER_SESSION_ID': 'test_session_id',
                     'SESSION_INDEX': 0,
                 })
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'get-history-video-ids/00-empty.json').read_text()))
    result = list(client.get_history_video_ids(return_dict=False))
    assert result == []


def test_get_history_video_ids_missing_video_id(mocker: MockerFixture, requests_mock: Mocker,
                                                client: YouTubeClient, data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'VISITOR_DATA': 'test_visitor_data',
                     'USER_SESSION_ID': 'test_session_id',
                     'SESSION_INDEX': 0,
                 })
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'get-history-video-ids/00-missing-video-id.json').read_text()))
    result = list(client.get_history_video_ids(return_dict=False))
    assert result == []



from __future__ import annotations

from typing import TYPE_CHECKING
import json

from youtube_unofficial.constants import WATCH_HISTORY_URL

if TYPE_CHECKING:
    from pathlib import Path

    from _pytest.logging import LogCaptureFixture
    from pytest_mock import MockerFixture
    from requests_mock import Mocker
    from youtube_unofficial.client import YouTubeClient


def test_get_history_info_no_continuation(mocker: MockerFixture, requests_mock: Mocker,
                                          client: YouTubeClient, data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'USER_SESSION_ID': 'test_session_id',
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'SESSION_INDEX': 0,
                     'VISITOR_DATA': 'test_visitor_data',
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'get-history-info/00-no-continuation.json').read_text()))
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    result = list(client.get_history_info())
    assert result == [{'videoRenderer': {'videoId': 'test_video'}}]


def test_get_history_info_with_continuation(mocker: MockerFixture, requests_mock: Mocker,
                                            client: YouTubeClient, data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'USER_SESSION_ID': 'test_session_id',
                     'INNERTUBE_CONTEXT_CLIENT_VERSION': '1.0',
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'SESSION_INDEX': 0,
                     'VISITOR_DATA': 'test_visitor_data',
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'get-history-info/00-with-continuation.json').read_text()))
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    requests_mock.post(
        'https://www.youtube.com/youtubei/v1/browse',
        json=json.loads(
            (data_path / 'get-history-info/00-with-continuation-response.json').read_text()))
    result = list(client.get_history_info())
    assert result == [{
        'videoRenderer': {
            'videoId': 'test_video'
        }
    }, {
        'videoRenderer': {
            'videoId': 'test_video_2'
        }
    }]


def test_get_history_info_alt_continuation(mocker: MockerFixture, requests_mock: Mocker,
                                           client: YouTubeClient, data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'USER_SESSION_ID': 'test_session_id',
                     'INNERTUBE_CONTEXT_CLIENT_VERSION': '1.0',
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'SESSION_INDEX': 0,
                     'VISITOR_DATA': 'test_visitor_data',
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'get-history-info/00-alt-continuation.json').read_text()))
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    requests_mock.post(
        'https://www.youtube.com/youtubei/v1/browse',
        response_list=json.loads(
            (data_path / 'get-history-info/00-alt-continuation-response.json').read_text()))
    result = list(client.get_history_info())
    assert result == [{'videoRenderer': {'videoId': 'test_video'}}]


def test_get_history_info_no_continuation_on_2nd_req(mocker: MockerFixture, requests_mock: Mocker,
                                                     client: YouTubeClient,
                                                     caplog: LogCaptureFixture,
                                                     data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'USER_SESSION_ID': 'test_session_id',
                     'INNERTUBE_CONTEXT_CLIENT_VERSION': '1.0',
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'SESSION_INDEX': 0,
                     'VISITOR_DATA': 'test_visitor_data',
                 })
    mocker.patch(
        'youtube_unofficial.client.initial_data',
        return_value=json.loads(
            (data_path / 'get-history-info/00-no-continuation-on-2nd-req.json').read_text()))
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    requests_mock.post(
        'https://www.youtube.com/youtubei/v1/browse',
        response_list=json.loads(
            (data_path /
             'get-history-info/00-no-continuation-on-2nd-req-responses.json').read_text()))
    with caplog.at_level('INFO'):
        result = list(client.get_history_info())
        assert result == [{'videoRenderer': {'videoId': 'test_video'}}]
        assert 'end of watch history.' in caplog.records[0].message


def test_get_history_info_bad_continuation(mocker: MockerFixture, requests_mock: Mocker,
                                           client: YouTubeClient, data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'USER_SESSION_ID': 'test_session_id',
                     'INNERTUBE_CONTEXT_CLIENT_VERSION': '1.0',
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'SESSION_INDEX': 0,
                     'VISITOR_DATA': 'test_visitor_data',
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'get-history-info/00-bad-continuation.json').read_text()))
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    requests_mock.post('https://www.youtube.com/youtubei/v1/browse',
                       json={'onResponseReceivedActions': [{
                           'appendContinuationItemsAction': {}
                       }]})
    result = list(client.get_history_info())
    assert result == [{'videoRenderer': {'videoId': 'test_video'}}]


def test_get_history_info_no_videos(mocker: MockerFixture, requests_mock: Mocker,
                                    client: YouTubeClient, data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'USER_SESSION_ID': 'test_session_id',
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'SESSION_INDEX': 0,
                     'VISITOR_DATA': 'test_visitor_data',
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'get-history-info/00-no-videos.json').read_text()))
    requests_mock.get(WATCH_HISTORY_URL, text='<html></html>')
    result = list(client.get_history_info())
    assert result == []



"""Typed dictionaries for YouTube history data."""
from __future__ import annotations

from typing import TYPE_CHECKING

from typing_extensions import TypedDict

if TYPE_CHECKING:
    from collections.abc import Sequence

__all__ = ('HistoryVideoIDsEntry',)


class Text(TypedDict):
    """Text information."""
    text: str
    """Text content."""


class DescriptionSnippet(TypedDict):
    """Description snippet information."""
    runs: Sequence[Text]
    """Text runs."""


class MetadataBadgeRender(TypedDict):
    """Child of :py:class:`MetadataBadgeRendererTopDict`."""
    style: str
    """Badge style."""


class MetadataBadgeRendererTop(TypedDict):
    """Information about the badge displayed for the user (e.g. checkmark meaning 'verified')."""
    metadataBadgeRenderer: MetadataBadgeRender
    """Inner data."""


class HistoryVideoIDsEntry(TypedDict, total=False):
    """History entry information."""
    description: str
    """Video description."""
    long_byline_text: str
    """Long byline text. Usually video description but may be different."""
    owner_text: str
    """Owner text. Usually the channel name."""
    short_byline_text: str
    """Short byline text. Often channel name but may be different."""
    short_view_count_text: str
    """Short view count text."""
    video_id: str
    """Video ID."""
    view_count_text: str
    """View count text."""



from __future__ import annotations

from typing import TYPE_CHECKING
import json

if TYPE_CHECKING:
    from pathlib import Path

    from _pytest.logging import LogCaptureFixture
    from pytest_mock import MockerFixture
    from requests_mock import Mocker
    from youtube_unofficial.client import YouTubeClient


def test_clear_playlist_empty(mocker: MockerFixture, requests_mock: Mocker, client: YouTubeClient,
                              caplog: LogCaptureFixture, data_path: Path) -> None:
    requests_mock.get('https://www.youtube.com/playlist?list=test_playlist', text='<html></html>')
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'VISITOR_DATA': 'test_visitor_data',
                     'USER_SESSION_ID': 'test_session_id',
                     'SESSION_INDEX': 0,
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads((data_path / 'clear-playlist/00-empty.json').read_text()))
    with caplog.at_level('INFO'):
        client.clear_playlist(playlist_id='test_playlist')
        assert 'playlist is empty.' in caplog.records[0].message


def test_clear_playlist_empty_iter(mocker: MockerFixture, requests_mock: Mocker,
                                   client: YouTubeClient, data_path: Path) -> None:
    requests_mock.get('https://www.youtube.com/playlist?list=test_playlist', text='<html></html>')
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'VISITOR_DATA': 'test_visitor_data',
                     'USER_SESSION_ID': 'test_session_id',
                     'SESSION_INDEX': 0,
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'clear-playlist/00-empty-iter.json').read_text()))
    rem_video_id = mocker.spy(client, 'remove_video_id_from_playlist')
    client.clear_playlist(playlist_id='test_playlist')
    assert rem_video_id.call_count == 0


def test_clear_playlist(mocker: MockerFixture, requests_mock: Mocker, client: YouTubeClient,
                        data_path: Path) -> None:
    requests_mock.get('https://www.youtube.com/playlist?list=test_playlist', text='<html></html>')
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'VISITOR_DATA': 'test_visitor_data',
                     'USER_SESSION_ID': 'test_session_id',
                     'SESSION_INDEX': 0,
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads((data_path / 'clear-playlist/00.json').read_text()))
    mocker.patch.object(client, 'remove_video_id_from_playlist')
    rem_video_id = mocker.spy(client, 'remove_video_id_from_playlist')
    client.clear_playlist(playlist_id='test_playlist')
    assert rem_video_id.call_count == 2



"""Commands."""
from __future__ import annotations

from typing import TYPE_CHECKING
import json
import logging

from bascom import setup_logging
import click

from . import YouTubeClient

if TYPE_CHECKING:
    from collections.abc import Iterable

__all__ = ('clear_watch_history', 'clear_watch_later', 'print_history', 'print_playlist',
           'print_watch_later', 'remove_history_entries', 'remove_video_id',
           'remove_watch_later_video_id', 'toggle_watch_history')


def print_playlist_ids_callback(browser: str,
                                profile: str,
                                playlist_id: str,
                                *,
                                output_json: bool = False) -> None:
    yt = YouTubeClient(browser, profile)
    for entry in yt.get_playlist_video_ids(playlist_id,
                                           return_dict=output_json):  # type: ignore[call-overload]
        if output_json:
            click.echo(json.dumps(entry, sort_keys=True))
        else:
            click.echo(entry)


@click.command(context_settings={'help_option_names': ('-h', '--help')})
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-b', '--browser', default='chrome', help='Browser to read cookies from.')
@click.option('-p', '--profile', default='Default', help='Browser profile.')
@click.option('-j', '--json', 'output_json', is_flag=True, help='Output in JSON format.')
def print_watch_later(browser: str,
                      profile: str,
                      *,
                      debug: bool = False,
                      output_json: bool = False) -> None:
    """
    Print your Watch Later playlist.

    By default, this will print the video IDs of your Watch Later playlist.

    If -j/--json is specified, this will print a JSON object for each video in your Watch Later
    playlist. The JSON object will have the following interface:

    \b

    .. code-block:: typescript

       {
           owner: string
           title: string
           video_id: string
           watch_url: string
       }
    """  # noqa: D301
    setup_logging(debug=debug,
                  loggers={
                      'youtube_unofficial': {
                          'handlers': ('console',),
                          'level': logging.DEBUG if debug else logging.INFO,
                          'propagate': False,
                      }
                  })
    print_playlist_ids_callback(browser, profile, 'WL', output_json=output_json)


@click.command(context_settings={'help_option_names': ('-h', '--help')})
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-b', '--browser', default='chrome', help='Browser to read cookies from.')
@click.option('-p', '--profile', default='Default', help='Browser profile.')
@click.option('-j', '--json', 'output_json', is_flag=True, help='Output in JSON format.')
@click.argument('playlist_id')
def print_playlist(browser: str,
                   profile: str,
                   playlist_id: str,
                   *,
                   debug: bool = False,
                   output_json: bool = False) -> None:
    """
    Print a playlist.

    By default, this will print the video IDs of the given playlist ID.

    If -j/--json is specified, this will print a JSON object for each video in your Watch Later
    playlist. The JSON object will have the following interface:

    \b

    .. code-block:: typescript

       {
           owner: string
           title: string
           video_id: string
           watch_url: string
       }
    """  # noqa: D301
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO,
                        format='%(levelname)s:%(name)s:%(lineno)d:%(funcName)s:%(message)s')
    print_playlist_ids_callback(browser, profile, playlist_id, output_json=output_json)


@click.command(context_settings={'help_option_names': ('-h', '--help')})
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-b', '--browser', default='chrome', help='Browser to read cookies from.')
@click.option('-p', '--profile', default='Default', help='Browser profile.')
@click.option('-j', '--json', 'output_json', is_flag=True, help='Output in JSON format.')
def print_history(browser: str,
                  profile: str,
                  *,
                  debug: bool = False,
                  output_json: bool = False) -> None:
    """Print your watch history.

    By default, this will print the video IDs of your watch history.

    If -j/--json is specified, this will print a JSON object for each video in your watch
    history. The JSON object will have the following interface:

    \b

    .. code-block:: typescript

       {
           channel_thumbnails: {
               height: int
               width: int
               url: string
           }[]
           description: string
           length: string
           length_accessible: string
           long_byline_text: string
           moving_thumbnails: {
               height: int
               width: int
               url: string
           }[]
           owner_text: string
           short_byline_text: string
           short_view_count_text: string
           title: string
           video_id: string
           video_thumbnails: {
               height: int
               width: int
               url: string
           }[]
        }
    """  # noqa: D301
    setup_logging(debug=debug,
                  loggers={
                      'youtube_unofficial': {
                          'handlers': ('console',),
                          'level': logging.DEBUG if debug else logging.INFO,
                          'propagate': False,
                      }
                  })
    yt = YouTubeClient(browser, profile)
    for entry in yt.get_history_video_ids(return_dict=output_json):  # type: ignore[call-overload]
        if output_json:
            click.echo(json.dumps(entry, sort_keys=True))
        else:
            click.echo(entry)


@click.command(context_settings={'help_option_names': ('-h', '--help')})
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-b', '--browser', default='chrome', help='Browser to read cookies from.')
@click.option('-p', '--profile', default='Default', help='Browser profile.')
@click.argument('video_ids', nargs=-1)
def remove_history_entries(browser: str,
                           profile: str,
                           video_ids: tuple[str, ...],
                           *,
                           debug: bool = False) -> None:
    """Remove videos from Watch History."""
    setup_logging(debug=debug,
                  loggers={
                      'youtube_unofficial': {
                          'handlers': ('console',),
                          'level': logging.DEBUG if debug else logging.INFO,
                          'propagate': False,
                      }
                  })
    yt = YouTubeClient(browser, profile)
    for video_id in video_ids:
        yt.remove_video_ids_from_history(video_id)


def remove_svi_callback(browser: str, profile: str, playlist_id: str,
                        video_ids: Iterable[str]) -> None:
    yt = YouTubeClient(browser, profile)
    for svi in video_ids:
        yt.remove_video_id_from_playlist(playlist_id, svi)


@click.command(context_settings={'help_option_names': ('-h', '--help')})
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-b', '--browser', default='chrome', help='Browser to read cookies from.')
@click.option('-p', '--profile', default='Default', help='Browser profile.')
@click.argument('video_ids', nargs=-1)
def remove_watch_later_video_id(browser: str,
                                profile: str,
                                video_ids: tuple[str, ...],
                                *,
                                debug: bool = False) -> None:
    """Remove videos from your Watch Later queue."""
    setup_logging(debug=debug,
                  loggers={
                      'youtube_unofficial': {
                          'handlers': ('console',),
                          'level': logging.DEBUG if debug else logging.INFO,
                          'propagate': False,
                      }
                  })
    remove_svi_callback(browser, profile, 'WL', video_ids)


@click.command(context_settings={'help_option_names': ('-h', '--help')})
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-b', '--browser', default='chrome', help='Browser to read cookies from.')
@click.option('-p', '--profile', default='Default', help='Browser profile.')
@click.argument('playlist_id', nargs=1)
@click.argument('video_ids', nargs=-1)
def remove_video_id(browser: str,
                    profile: str,
                    playlist_id: str,
                    video_ids: tuple[str, ...],
                    *,
                    debug: bool = False) -> None:
    """Remove videos from a playlist."""
    setup_logging(debug=debug,
                  loggers={
                      'youtube_unofficial': {
                          'handlers': ('console',),
                          'level': logging.DEBUG if debug else logging.INFO,
                          'propagate': False,
                      }
                  })
    remove_svi_callback(browser, profile, playlist_id, video_ids)


@click.command(context_settings={'help_option_names': ('-h', '--help')})
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-b', '--browser', default='chrome', help='Browser to read cookies from.')
@click.option('-p', '--profile', default='Default', help='Browser profile.')
def toggle_watch_history(browser: str, profile: str, *, debug: bool = False) -> None:
    """Disable or enable watch history."""  # noqa: DOC501
    setup_logging(debug=debug,
                  loggers={
                      'youtube_unofficial': {
                          'handlers': ('console',),
                          'level': logging.DEBUG if debug else logging.INFO,
                          'propagate': False,
                      }
                  })
    if not YouTubeClient(browser=browser, profile=profile).toggle_watch_history():
        click.echo('Failed to toggle watch history.', err=True)
        raise click.Abort
    click.echo('Watch history toggled.')


@click.command(context_settings={'help_option_names': ('-h', '--help')})
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-b', '--browser', default='chrome', help='Browser to read cookies from.')
@click.option('-p', '--profile', default='Default', help='Browser profile.')
def clear_watch_history(browser: str, profile: str, *, debug: bool = False) -> None:
    """Clear watch history."""
    setup_logging(debug=debug,
                  loggers={
                      'youtube_unofficial': {
                          'handlers': ('console',),
                          'level': logging.DEBUG if debug else logging.INFO,
                          'propagate': False,
                      }
                  })
    yt = YouTubeClient(browser, profile)
    yt.clear_watch_history()
    click.echo('Watch history cleared.')


@click.command(context_settings={'help_option_names': ('-h', '--help')})
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-b', '--browser', default='chrome', help='Browser to read cookies from.')
@click.option('-p', '--profile', default='Default', help='Browser profile.')
def clear_watch_later(browser: str, profile: str, *, debug: bool = False) -> None:
    """Clear watch later queue."""
    setup_logging(debug=debug,
                  loggers={
                      'youtube_unofficial': {
                          'handlers': ('console',),
                          'level': logging.DEBUG if debug else logging.INFO,
                          'propagate': False,
                      }
                  })
    yt = YouTubeClient(browser, profile)
    yt.clear_watch_later()
    click.echo('Watch later queue cleared.')



"""Tests that have no initial data."""
from __future__ import annotations

from typing import TYPE_CHECKING

from youtube_unofficial.constants import WATCH_LATER_URL

if TYPE_CHECKING:
    from pytest_mock import MockerFixture
    from requests_mock import Mocker
    from youtube_unofficial.client import YouTubeClient


def test_remove_video_id_from_playlist(mocker: MockerFixture, requests_mock: Mocker,
                                       client: YouTubeClient) -> None:
    ytcfg_mock = {
        'INNERTUBE_API_KEY': 'test_api_key',
        'VISITOR_DATA': 'test_visitor_data',
        'USER_SESSION_ID': 'test_session_id',
        'SESSION_INDEX': 0,
    }
    mocker.patch('youtube_unofficial.client.find_ytcfg', return_value=ytcfg_mock)
    mocker.patch('youtube_unofficial.client.context_client_body', return_value={})
    mocker.patch('youtube_unofficial.client.Soup')

    requests_mock.get(WATCH_LATER_URL, text='<html></html>')
    requests_mock.post(
        'https://www.youtube.com/youtubei/v1/browse/edit_playlist',
        json={'status': 'STATUS_SUCCEEDED'},
    )

    result = client.remove_video_id_from_playlist(playlist_id='test_playlist',
                                                  video_id='test_video')
    assert result is True


def test_clear_playlist(mocker: MockerFixture, client: YouTubeClient) -> None:
    mock_session = mocker.Mock()
    mocker.patch('yt_dlp_utils.setup_session', return_value=mock_session)
    mocker.patch.object(client, 'get_playlist_info', return_value=iter([]))
    client.clear_playlist(playlist_id='test_playlist')


def test_clear_watch_later(mocker: MockerFixture, client: YouTubeClient) -> None:
    mock_session = mocker.Mock()
    mocker.patch('yt_dlp_utils.setup_session', return_value=mock_session)
    mocker.patch.object(client, 'clear_playlist')
    client.clear_watch_later()
    client.clear_playlist.assert_called_once_with('WL')  # type: ignore[attr-defined]


def test_remove_video_ids_from_history_no_entries(mocker: MockerFixture,
                                                  client: YouTubeClient) -> None:
    mock_session = mocker.MagicMock()
    mocker.patch('yt_dlp_utils.setup_session', return_value=mock_session)
    mocker.patch.object(client, 'get_history_info', return_value=iter([]))
    mocker.patch('youtube_unofficial.client.Soup')
    mocker.patch('youtube_unofficial.client.find_ytcfg', return_value={})
    result = client.remove_video_ids_from_history(video_ids=['test_video'])
    assert result is False


def test_remove_video_ids_from_history_empty(mocker: MockerFixture, requests_mock: Mocker,
                                             client: YouTubeClient) -> None:
    result = client.remove_video_ids_from_history([])
    assert result is False


def test_remove_video_from_playlist_cached(mocker: MockerFixture, requests_mock: Mocker,
                                           client: YouTubeClient) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'USER_SESSION_ID': 'test_session_id',
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'SESSION_INDEX': 0,
                     'VISITOR_DATA': 'test_visitor_data',
                 })
    mocker.patch('youtube_unofficial.client.context_client_body', return_value={})
    mocker.patch('youtube_unofficial.client.Soup')

    requests_mock.get(WATCH_LATER_URL, text='<html></html>')
    requests_mock.post(
        'https://www.youtube.com/youtubei/v1/browse/edit_playlist',
        json={'status': 'STATUS_SUCCEEDED'},
    )

    result = client.remove_video_id_from_playlist('test_playlist', 'test_video', cache_values=True)
    assert result is True
    result = client.remove_video_id_from_playlist('test_playlist', 'test_video', cache_values=True)
    assert result is True
    assert len(requests_mock.request_history) == 3


def test_remove_set_video_id_from_playlist(mocker: MockerFixture, requests_mock: Mocker,
                                           client: YouTubeClient) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'USER_SESSION_ID': 'test_session_id',
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'SESSION_INDEX': 0,
                     'VISITOR_DATA': 'test_visitor_data',
                 })
    mocker.patch('youtube_unofficial.client.context_client_body', return_value={})
    mocker.patch('youtube_unofficial.client.Soup')

    requests_mock.get(WATCH_LATER_URL, text='<html></html>')
    requests_mock.post(
        'https://www.youtube.com/youtubei/v1/browse/edit_playlist',
        json={'status': 'STATUS_SUCCEEDED'},
    )

    result = client.remove_set_video_id_from_playlist('test_playlist', 'test_video')
    assert result is True


def test_remove_set_video_id_from_playlist_cached(mocker: MockerFixture, requests_mock: Mocker,
                                                  client: YouTubeClient) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'USER_SESSION_ID': 'test_session_id',
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'SESSION_INDEX': 0,
                     'VISITOR_DATA': 'test_visitor_data',
                 })
    mocker.patch('youtube_unofficial.client.context_client_body', return_value={})
    mocker.patch('youtube_unofficial.client.Soup')

    requests_mock.get(WATCH_LATER_URL, text='<html></html>')
    requests_mock.post(
        'https://www.youtube.com/youtubei/v1/browse/edit_playlist',
        json={'status': 'STATUS_SUCCEEDED'},
    )

    result = client.remove_set_video_id_from_playlist('test_playlist',
                                                      'test_video',
                                                      cache_values=True)
    assert result is True
    result = client.remove_set_video_id_from_playlist('test_playlist',
                                                      'test_video',
                                                      cache_values=True)
    assert result is True
    assert len(requests_mock.request_history) == 3



"""Client library."""
from __future__ import annotations

from collections.abc import Iterable, Mapping
from datetime import datetime, timezone
from operator import itemgetter
from typing import TYPE_CHECKING, Any, Literal, cast
import hashlib
import json
import logging

from bs4 import BeautifulSoup as Soup
from typing_extensions import overload
import yt_dlp_utils

from .constants import (
    EXTRACTED_THUMBNAIL_KEYS,
    HISTORY_ENTRY_KEYS_TO_SKIP,
    SIMPLE_TEXT_KEYS,
    TEXT_RUNS_KEYS,
    USER_AGENT,
    WATCH_HISTORY_URL,
    WATCH_LATER_URL,
)
from .download import download_page
from .typing.playlist import PlaylistVideoIDsEntry
from .utils import (
    context_client_body,
    extract_keys,
    find_ytcfg,
    get_text_runs,
    initial_data,
    ytcfg_headers,
)

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from .typing.history import (
        DescriptionSnippet,
        HistoryVideoIDsEntry,
        MetadataBadgeRendererTop,
    )
    from .typing.playlist import PlaylistInfo, PlaylistVideoListRenderer
    from .typing.ytcfg import YtcfgDict

__all__ = ('NoFeedbackToken', 'YouTubeClient')

log = logging.getLogger(__name__)


class NoFeedbackToken(Exception):
    """No feedback token found."""
    def __init__(self) -> None:
        super().__init__('No feedback token found.')


class YouTubeClient:
    """YouTube client for managing playlists and history."""
    def __init__(self, browser: str, profile: str) -> None:
        """
        Initialise the client.

        Parameters
        ----------
        browser : str
            The browser to use.
        profile : str
            The profile to use.
        """
        self.session = yt_dlp_utils.setup_session(browser,
                                                  profile,
                                                  domains={'.youtube.com'},
                                                  setup_retry=True)
        """Requests :py:class:`requests.Session` instance."""
        self.session.headers.update({'User-Agent': USER_AGENT})
        self._rsvi_cache: dict[str, Any] | None = None

    def remove_video_id_from_playlist(self,
                                      playlist_id: str,
                                      video_id: str,
                                      *,
                                      cache_values: bool | None = False) -> bool:
        """
        Remove a video from a playlist.

        Parameters
        ----------
        playlist_id : str
            The ID of the playlist.
        video_id : str
            The ID of the video to remove.
        cache_values : bool | None
            If ``True``, cache the values of the page and ytcfg. This is useful for performance
            reasons, but may cause issues if updates are needed.

        Returns
        -------
        bool
            ``True`` if the operation was successful, ``False`` otherwise.
        """
        if cache_values and self._rsvi_cache:
            soup = self._rsvi_cache['soup']
            ytcfg = self._rsvi_cache['ytcfg']
            headers = self._rsvi_cache['headers']
        else:
            soup = self._download_page_soup(WATCH_LATER_URL)
            ytcfg = find_ytcfg(soup)
            headers = ytcfg_headers(ytcfg)
        if cache_values:
            self._rsvi_cache = {'soup': soup, 'ytcfg': ytcfg, 'headers': headers}
        action = {'removedVideoId': video_id, 'action': 'ACTION_REMOVE_VIDEO_BY_VIDEO_ID'}
        assert 'INNERTUBE_API_KEY' in ytcfg
        assert 'VISITOR_DATA' in ytcfg
        assert 'DELEGATED_SESSION_ID' in ytcfg or 'USER_SESSION_ID' in ytcfg
        delegated_session_id = ytcfg.get('DELEGATED_SESSION_ID')
        session_index = ytcfg.get('SESSION_INDEX')
        return cast(
            'bool',
            self._download_page(
                'https://www.youtube.com/youtubei/v1/browse/edit_playlist',
                method='post',
                params={'key': ytcfg['INNERTUBE_API_KEY']},
                headers={
                    'Authorization': self._authorization_sapisidhash_header(),
                    'x-goog-authuser': f'{session_index or 0}',
                    'x-origin': 'https://www.youtube.com',
                    'x-goog-visitor-id': ytcfg['VISITOR_DATA'],
                    'accept': '*/*',
                    'origin': 'https://www.youtube.com',
                    'referer': f'https://www.youtube.com/playlist?list={playlist_id}',
                } | ({
                    'x-goog-pageid': delegated_session_id
                } if delegated_session_id else {}),
                json={
                    'actions': [action],
                    'playlistId': playlist_id,
                    'params': 'CAFAAQ%3D%3D',
                    'context': {
                        'client': context_client_body(ytcfg),
                        'request': {
                            'consistencyTokenJars': [],
                            'internalExperimentFlags': []
                        },
                    }
                },
                return_json=True)['status'] == 'STATUS_SUCCEEDED')

    def remove_set_video_id_from_playlist(self,
                                          playlist_id: str,
                                          set_video_id: str,
                                          *,
                                          cache_values: bool | None = False) -> bool:
        """
        Remove a video from a playlist by its *setVideoId*.

        Parameters
        ----------
        playlist_id : str
            The ID of the playlist.
        set_video_id : str
            The *setVideoId* of the video to remove.
        cache_values : bool | None
            If ``True``, cache the values of the page and ytcfg. This is useful for performance
            reasons, but may cause issues if updates are needed.

        Returns
        -------
        bool
            ``True`` if the operation was successful, ``False`` otherwise.
        """
        if cache_values and self._rsvi_cache:
            soup = self._rsvi_cache['soup']
            ytcfg = self._rsvi_cache['ytcfg']
            headers = self._rsvi_cache['headers']
        else:
            soup = self._download_page_soup(WATCH_LATER_URL)
            ytcfg = find_ytcfg(soup)
            headers = ytcfg_headers(ytcfg)
        if cache_values:
            self._rsvi_cache = {'soup': soup, 'ytcfg': ytcfg, 'headers': headers}
        assert 'INNERTUBE_API_KEY' in ytcfg
        assert 'VISITOR_DATA' in ytcfg
        assert 'DELEGATED_SESSION_ID' in ytcfg or 'USER_SESSION_ID' in ytcfg
        return cast(
            'bool',
            self._download_page(
                'https://www.youtube.com/youtubei/v1/browse/edit_playlist',
                method='post',
                params={'key': ytcfg['INNERTUBE_API_KEY']},
                headers={
                    'Authorization': self._authorization_sapisidhash_header(),
                    'x-goog-authuser': f'{ytcfg.get("SESSION_INDEX", 0)}',
                    'x-origin': 'https://www.youtube.com',
                    'x-goog-visitor-id': ytcfg['VISITOR_DATA'],
                    'accept': '*/*',
                    'origin': 'https://www.youtube.com',
                    'referer': f'https://www.youtube.com/playlist?list={playlist_id}',
                } | ({
                    'x-goog-pageid': ytcfg['DELEGATED_SESSION_ID']
                } if ytcfg.get('DELEGATED_SESSION_ID') else {}),
                json={
                    'actions': [{
                        'action': 'ACTION_REMOVE_VIDEO',
                        'setVideoId': set_video_id,
                    }],
                    'playlistId': playlist_id,
                    'params': 'CAFAAQ%3D%3D',
                    'context': {
                        'client': context_client_body(ytcfg),
                        'request': {
                            'consistencyTokenJars': [],
                            'internalExperimentFlags': [],
                            'useSsl': True
                        },
                        'user': {
                            'lockedSafetyMode': False
                        }
                    }
                },
                return_json=True)['status'] == 'STATUS_SUCCEEDED')

    def clear_watch_history(self) -> bool:
        """
        Clear watch history.

        Returns
        -------
        bool
            ``True`` if the operation was successful, ``False`` otherwise.

        Raises
        ------
        NoFeedbackToken
        """
        content = self._download_page_soup(WATCH_HISTORY_URL)
        ytcfg = find_ytcfg(content)
        init_data = initial_data(content)
        # If there are no videos, there is no search and index is 0.
        browse_feed_actions_renderer = (init_data['contents']['twoColumnBrowseResultsRenderer']
                                        ['secondaryContents']['browseFeedActionsRenderer'])
        try:
            if browse_feed_actions_renderer['contents'][0]['buttonRenderer']['isDisabled']:
                log.debug('Clear history button is disabled.')
                return False
        except (IndexError, KeyError):
            pass
        try:
            feedback_token = (
                browse_feed_actions_renderer['contents'][1]['buttonRenderer']['navigationEndpoint']
                ['confirmDialogEndpoint']['content']['confirmDialogRenderer']['confirmEndpoint']
                ['feedbackEndpoint']['feedbackToken'])
        except (IndexError, KeyError) as e:
            raise NoFeedbackToken from e
        return cast('bool', self._single_feedback_api_call(ytcfg, feedback_token))

    def get_playlist_info(self, playlist_id: str) -> Iterator[PlaylistInfo]:
        """
        Get playlist information.

        Parameters
        ----------
        playlist_id : str
            The ID of the playlist.

        Yields
        ------
        PlaylistInfo
            The playlist information.

        Raises
        ------
        KeyError
        """
        url = f'https://www.youtube.com/playlist?list={playlist_id}'
        content = self._download_page_soup(url)
        ytcfg = find_ytcfg(content)
        ytcfg_headers(ytcfg)
        yt_init_data = initial_data(content)
        video_list_renderer: PlaylistVideoListRenderer | None = None
        try:
            video_list_renderer = (
                yt_init_data['contents']['twoColumnBrowseResultsRenderer']['tabs'][0]['tabRenderer']
                ['content']['sectionListRenderer']['contents'][0]['itemSectionRenderer']['contents']
                [0]['playlistVideoListRenderer'])
        except KeyError as e:
            if e.args[0] == 'playlistVideoListRenderer':
                msg = 'This playlist might be empty.'
                raise KeyError(msg) from e
            raise
        assert video_list_renderer is not None
        try:
            for item in video_list_renderer['contents']:
                if 'playlistVideoRenderer' in item:
                    yield item
                elif 'continuationItemRenderer' in item:
                    break
        except KeyError:
            yield from []
        endpoint = continuation = api_url = None
        try:
            endpoint = (video_list_renderer['contents'][-1]['continuationItemRenderer']
                        ['continuationEndpoint'])
            api_url = endpoint['commandMetadata']['webCommandMetadata']['apiUrl']
            continuation = endpoint['continuationCommand']['token']
        except KeyError:
            pass
        if continuation and api_url:
            while True:
                contents = self._single_feedback_api_call(ytcfg,
                                                          api_url=api_url,
                                                          merge_json={'continuation': continuation},
                                                          return_is_processed=False)
                assert isinstance(contents, dict)
                assert 'onResponseReceivedActions' in contents
                items = contents['onResponseReceivedActions'][0]['appendContinuationItemsAction'][
                    'continuationItems']
                for item in items:
                    if 'playlistVideoRenderer' in item:
                        yield item
                    elif 'continuationItemRenderer' in item:
                        endpoint = item['continuationItemRenderer']['continuationEndpoint']
                        continuation = endpoint['continuationCommand']['token']
                        break
                if 'continuationItemRenderer' not in items[-1]:
                    break

    @overload
    def get_playlist_video_ids(self, playlist_id: str) -> Iterator[str]:  # pragma: no cover
        ...

    @overload
    def get_playlist_video_ids(
            self, playlist_id: str, *,
            return_dict: Literal[True]) -> Iterator[PlaylistVideoIDsEntry]:  # pragma: no cover
        ...

    @overload
    def get_playlist_video_ids(self, playlist_id: str, *,
                               return_dict: Literal[False]) -> Iterator[str]:  # pragma: no cover
        ...

    def get_playlist_video_ids(
            self,
            playlist_id: str,
            *,
            return_dict: bool = False) -> Iterator[str] | Iterator[PlaylistVideoIDsEntry]:
        """
        Get video IDs from a playlist.

        Parameters
        ----------
        playlist_id : str
            The ID of the playlist.
        return_dict : bool
            If ``True``, yield dictionaries.

        Yields
        ------
        str | PlaylistVideoIDsEntry
            The video IDs or dictionaries with video information.
        """
        for item in self.get_playlist_info(playlist_id):
            renderer = item['playlistVideoRenderer']
            if 'videoId' not in renderer:
                continue
            owner = title = None
            if 'shortBylineText' in renderer:
                if 'runs' in renderer['shortBylineText']:
                    owner = ' - '.join(map(itemgetter('text'), renderer['shortBylineText']['runs']))
                elif 'text' in renderer['shortBylineText']:
                    owner = renderer['shortBylineText']['text']
            if 'title' in renderer:
                if 'simpleText' in renderer['title']:
                    title = renderer['title']['simpleText']
                elif 'runs' in renderer['title']:
                    title = ' - '.join(map(itemgetter('text'), renderer['title']['runs']))
            if return_dict:
                yield PlaylistVideoIDsEntry(
                    owner=owner,
                    title=title,
                    video_id=renderer['videoId'],
                    watch_url=f'https://www.youtube.com/watch?v={renderer["videoId"]}')
            else:
                yield renderer['videoId']

    def clear_playlist(self, playlist_id: str) -> None:
        """
        Remove all videos from the specified playlist.

        Use `WL` for Watch Later.

        Parameters
        ----------
        playlist_id : str
            The ID of the playlist.
        """
        playlist_info = self.get_playlist_info(playlist_id)
        try:
            video_ids = [x['playlistVideoRenderer']['videoId'] for x in playlist_info]
        except KeyError:
            log.info('Caught KeyError. This probably means the playlist is empty.')
            return
        for video_id in video_ids:
            log.debug('Deleting from playlist: video_id = %s', video_id)
            self.remove_video_id_from_playlist(playlist_id, video_id)

    def clear_watch_later(self) -> None:
        """Remove all videos from the 'Watch Later' playlist."""
        self.clear_playlist('WL')

    def get_history_info(self) -> Iterator[dict[str, Any]]:
        """
        Get information about the History playlist.

        Yields
        ------
        dict[str, Any]
            The history information.
        """
        content = self._download_page_soup(WATCH_HISTORY_URL)
        init_data = initial_data(content)
        ytcfg = find_ytcfg(content)
        section_list_renderer = (init_data['contents']['twoColumnBrowseResultsRenderer']['tabs'][0]
                                 ['tabRenderer']['content']['sectionListRenderer'])
        next_continuation = None
        for section_list in section_list_renderer['contents']:
            try:
                yield from section_list['itemSectionRenderer']['contents']
            except KeyError:  # noqa: PERF203
                if 'continuationItemRenderer' in section_list:
                    next_continuation = {
                        'continuation': (section_list['continuationItemRenderer']
                                         ['continuationEndpoint']['continuationCommand']['token'])
                    }
                    break
        if not next_continuation:
            try:
                next_continuation = (
                    section_list_renderer['continuations'][0]['nextContinuationData'])
            except KeyError:
                return
        assert next_continuation is not None
        params = {
            'continuation': next_continuation['continuation'],
            'ctoken': next_continuation['continuation']
        }
        while True:
            resp = self._single_feedback_api_call(
                ytcfg,
                api_url='/youtubei/v1/browse',
                merge_json={'continuation': params['continuation']},
                return_is_processed=False)
            contents = cast('dict[str, Any]', resp)
            try:
                section_list_renderer = (contents['onResponseReceivedActions'][0]
                                         ['appendContinuationItemsAction']['continuationItems'])
            except KeyError as e:
                log.debug('Caught KeyError: %s. Possible keys: %s', e, ', '.join(contents.keys()))
                break
            continuations = None
            for section_list in section_list_renderer:
                try:
                    yield from section_list['itemSectionRenderer']['contents']
                except KeyError:  # noqa: PERF203
                    if 'continuationItemRenderer' in section_list:
                        continuations = {
                            'continuation': (
                                section_list['continuationItemRenderer']['continuationEndpoint']
                                ['continuationCommand']['token']),
                            'clickTrackingParams': (section_list['continuationItemRenderer']
                                                    ['continuationEndpoint']['clickTrackingParams'])
                        }
                    break
            if not continuations:
                log.info('Likely hit the end of watch history.')
                break
            params['itct'] = continuations['clickTrackingParams']
            params['ctoken'] = continuations['continuation']
            params['continuation'] = continuations['continuation']

    @overload
    def get_history_video_ids(self) -> Iterator[str]:  # pragma: no cover
        ...

    @overload
    def get_history_video_ids(
            self,
            *,
            return_dict: Literal[True] = True
    ) -> Iterator[HistoryVideoIDsEntry]:  # pragma: no cover
        ...

    @overload
    def get_history_video_ids(self,
                              *,
                              return_dict: Literal[False] = False
                              ) -> Iterator[str]:  # pragma: no cover
        ...

    def get_history_video_ids(  # noqa: C901
            self, *, return_dict: bool = False) -> Iterator[str] | Iterator[HistoryVideoIDsEntry]:
        """
        Get video IDs from the History playlist.

        Yields
        ------
        str
            The video IDs.
        """
        def is_verified(owner_badges: Iterable[MetadataBadgeRendererTop]) -> bool:
            for badge in (x['metadataBadgeRenderer'] for x in owner_badges):
                if badge['style'] == 'BADGE_STYLE_TYPE_VERIFIED':
                    return True
            return False

        for entry in self.get_history_info():
            d: dict[str, Any] = {}
            if 'videoId' not in entry.get('videoRenderer', {}):
                continue
            if return_dict:
                for k, v in sorted(entry.get('videoRenderer', {}).items()):
                    if k in HISTORY_ENTRY_KEYS_TO_SKIP:
                        continue
                    if k == 'videoId':
                        assert isinstance(v, str)
                        d['video_id'] = v
                    elif isinstance(v, int | str | float | bool):
                        d[k] = v
                    elif k in TEXT_RUNS_KEYS and isinstance(v, Mapping):
                        d[TEXT_RUNS_KEYS[k]] = get_text_runs(cast('DescriptionSnippet', v))
                    elif k == 'richThumbnail' and isinstance(v, Mapping):
                        if 'moving_thumbnails' not in d:  # pragma: no cover
                            d['moving_thumbnails'] = []
                        d['moving_thumbnails'] += [
                            extract_keys(EXTRACTED_THUMBNAIL_KEYS, t) for t in
                            v['movingThumbnailRenderer']['movingThumbnailDetails']['thumbnails']
                        ]
                    elif k == 'channelThumbnailSupportedRenderers' and isinstance(v, Mapping):
                        if 'video_thumbnails' not in d:  # pragma: no cover
                            d['video_thumbnails'] = []
                        d['video_thumbnails'] += [
                            extract_keys(EXTRACTED_THUMBNAIL_KEYS, t) for t in
                            v['channelThumbnailWithLinkRenderer']['thumbnail']['thumbnails']
                        ]
                    elif k in SIMPLE_TEXT_KEYS and isinstance(v, Mapping):
                        d[SIMPLE_TEXT_KEYS[k]] = v['simpleText']
                    elif k == 'lengthText' and isinstance(v, Mapping):
                        d['length_accessible'] = v['accessibility']['accessibilityData']['label']
                        d['length'] = v['simpleText']
                    elif k == 'ownerBadges' and isinstance(v, Iterable):
                        d['verified'] = is_verified(v)
                    elif k == 'thumbnail' and isinstance(v, Mapping):
                        if 'video_thumbnails' not in d:  # pragma: no cover
                            d['video_thumbnails'] = []
                        d['video_thumbnails'] += [
                            extract_keys(EXTRACTED_THUMBNAIL_KEYS, t) for t in v['thumbnails']
                        ]
                d['watch_url'] = f'https://www.youtube.com/watch?v={d["video_id"]}'
                yield cast('HistoryVideoIDsEntry', d)
            else:
                yield entry['videoRenderer']['videoId']

    def remove_video_ids_from_history(self, video_ids: Sequence[str]) -> bool:
        """
        Delete a history entry by video ID.

        Parameters
        ----------
        video_ids : Sequence[str]
            The video IDs to delete.

        Returns
        -------
        bool
            ``True`` if the operation was successful, ``False`` otherwise.
        """
        if not video_ids:
            return False
        history_info = self.get_history_info()
        content = self._download_page_soup(WATCH_HISTORY_URL)
        ytcfg = find_ytcfg(content)
        entries = [x for x in history_info if x['videoRenderer']['videoId'] in video_ids]
        if not entries:
            return False
        return all(
            self._single_feedback_api_call(
                ytcfg, entry['videoRenderer']['menu']['menuRenderer']['topLevelButtons'][0]
                ['buttonRenderer']['serviceEndpoint']['feedbackEndpoint']['feedbackToken'])
            for entry in entries)

    def _authorization_sapisidhash_header(self, ytcfg: YtcfgDict | None = None) -> str:
        now = int(datetime.now(timezone.utc).timestamp())
        sapisid = self.session.cookies.get(
            '__Secure-3PAPISID',
            self.session.cookies.get('SAPISID', self.session.cookies.get('__Secure-1PAPISID')))
        assert sapisid is not None
        if ytcfg:
            session_id = ytcfg['USER_SESSION_ID']
            # session_id = ytcfg.get('DELEGATED_SESSION_ID', ytcfg.get('USER_SESSION_ID'))  # noqa: E501, ERA001
            # assert session_id is not None
            m = hashlib.sha1(' '.join(  # noqa: S324
                (session_id, str(now), sapisid, 'https://www.youtube.com')).encode())
            a = '_'.join((str(now), m.hexdigest(), 'u'))
            return ' '.join(
                f'{type_} {a}' for type_ in ('SAPISIDHASH', 'SAPISID1PHASH', 'SAPISID3PHASH'))
        m = hashlib.sha1(f'{now} {sapisid} https://www.youtube.com'.encode())  # noqa: S324
        return f'SAPISIDHASH {now}_{m.hexdigest()}'

    def _single_feedback_api_call(self,
                                  ytcfg: YtcfgDict,
                                  feedback_token: str = '',
                                  api_url: str = '/youtubei/v1/feedback',
                                  merge_json: dict[str, Any] | None = None,
                                  click_tracking_params: str | None = None,
                                  *,
                                  return_is_processed: bool = True) -> dict[str, Any] | bool:
        if not merge_json:
            merge_json = {}
        feedback_token_part = {
            'feedbackTokens': [feedback_token],
            'isFeedbackTokenUnencrypted': False,
            'shouldMerge': False
        } if feedback_token else {}
        api_url = f'https://www.youtube.com{api_url}'
        log.debug('API URL: %s', api_url)
        assert 'DELEGATED_SESSION_ID' in ytcfg or 'USER_SESSION_ID' in ytcfg
        delegated_session_id = ytcfg.get('DELEGATED_SESSION_ID')
        headers = {
            'Authorization': self._authorization_sapisidhash_header(ytcfg),
            'x-goog-authuser': f'{ytcfg.get("SESSION_INDEX", 0)}',
            'x-origin': 'https://www.youtube.com',
            'X-Youtube-Bootstrap-Logged-In': 'true',
        } | ({
            'x-goog-pageid': delegated_session_id
        } if delegated_session_id else {})
        log.debug('Request headers: %s', headers)
        json_data = {
            'context': {
                'client': context_client_body(ytcfg),
            } | ({
                'clickTracking': {
                    'clickTrackingParams': click_tracking_params
                }
            } if click_tracking_params else {})
        } | feedback_token_part | merge_json
        log.debug('Request JSON: %s', json.dumps(json_data, indent=2, sort_keys=True))
        ret = self._download_page(api_url,
                                  method='post',
                                  params={'prettyPrint': 'false'},
                                  headers=headers,
                                  json=json_data,
                                  return_json=True)
        log.debug('Response JSON: %s', json.dumps(ret, indent=2, sort_keys=True))
        if ret.get('responseContext', {}).get('mainAppWebResponseContext', {}).get(
                'loggedOut', False):
            msg = ('Response indicates logged out. '
                   'Please check your cookies and try again.')
            raise RuntimeError(msg)
        if return_is_processed:
            try:
                return cast('bool', ret['feedbackResponses'][0]['isProcessed'])
            except KeyError:
                return False
        return ret

    def _toggle_history(self, page_url: str, contents_index: int) -> bool:
        content = self._download_page_soup(page_url)
        ytcfg = find_ytcfg(content)
        info = (initial_data(content)['contents']['twoColumnBrowseResultsRenderer']
                ['secondaryContents']['browseFeedActionsRenderer']['contents'][contents_index]
                ['buttonRenderer']['navigationEndpoint']['confirmDialogEndpoint']['content']
                ['confirmDialogRenderer']['confirmEndpoint'])
        return cast(
            'bool',
            self._single_feedback_api_call(ytcfg, info['feedbackEndpoint']['feedbackToken'],
                                           info['commandMetadata']['webCommandMetadata']['apiUrl']))

    def toggle_watch_history(self) -> bool:
        """
        Pauses or resumes watch history depending on the current state.

        Returns
        -------
        bool
            ``True`` if the operation was successful, ``False`` otherwise.
        """
        return self._toggle_history(WATCH_HISTORY_URL, 2)

    @overload
    def _download_page(self,
                       url: str,
                       data: Any = None,
                       method: Literal['get', 'post'] = 'get',
                       headers: Mapping[str, str] | None = None,
                       params: Mapping[str, str] | None = None,
                       json: Any = None,
                       *,
                       return_json: Literal[False] = False) -> str:  # pragma: no cover
        ...

    @overload
    def _download_page(self,
                       url: str,
                       data: Any = None,
                       method: Literal['get', 'post'] = 'get',
                       headers: Mapping[str, str] | None = None,
                       params: Mapping[str, str] | None = None,
                       json: Any = None,
                       *,
                       return_json: Literal[True]) -> dict[str, Any]:  # pragma: no cover
        ...

    def _download_page(self,
                       url: str,
                       data: Any = None,
                       method: Literal['get', 'post'] = 'get',
                       headers: Mapping[str, str] | None = None,
                       params: Mapping[str, str] | None = None,
                       json: Any = None,
                       *,
                       return_json: bool = False) -> str | dict[str, Any]:
        return download_page(  # type: ignore[call-overload,no-any-return]
            self.session,
            url,
            data,
            method,
            headers,
            params,
            json,
            return_json=return_json)

    def _download_page_soup(self, *args: Any, **kwargs: Any) -> Soup:
        return Soup(cast('str', self._download_page(*args, **kwargs)),
                    kwargs.pop('parser', 'html5lib'))



from __future__ import annotations

from typing import TYPE_CHECKING

from youtube_unofficial.main import main
import pytest

if TYPE_CHECKING:
    from click.testing import CliRunner
    from pytest_mock import MockerFixture


@pytest.fixture
def mock_setup_session(mocker: MockerFixture) -> None:
    mocker.patch('youtube_unofficial.client.yt_dlp_utils.setup_session')


def test_main_help(runner: CliRunner, mock_setup_session: None) -> None:
    result = runner.invoke(main, ['--help'])
    assert result.exit_code == 0
    assert '-h, --help' in result.output


def test_print_watch_later(mocker: MockerFixture, runner: CliRunner,
                           mock_setup_session: None) -> None:
    mocker.patch('youtube_unofficial.client.YouTubeClient.get_playlist_video_ids',
                 return_value=['1234'])
    result = runner.invoke(main, ['print-watch-later'])
    assert result.exit_code == 0
    assert '1234\n' in result.output


def test_print_watch_later_json(mocker: MockerFixture, runner: CliRunner,
                                mock_setup_session: None) -> None:
    mocker.patch('youtube_unofficial.client.YouTubeClient.get_playlist_video_ids',
                 return_value=[{
                     'z': 1,
                     'fake': 'object'
                 }])
    result = runner.invoke(main, ['print-watch-later', '--json'])
    assert result.exit_code == 0
    assert '{"fake": "object", "z": 1}' in result.output


def test_print_playlist(mocker: MockerFixture, runner: CliRunner, mock_setup_session: None) -> None:
    mocker.patch('youtube_unofficial.client.YouTubeClient.get_playlist_video_ids',
                 return_value=['1234'])
    result = runner.invoke(main, ['print-playlist', '1'])
    assert result.exit_code == 0
    assert '1234\n' in result.output


def test_print_playlist_json(mocker: MockerFixture, runner: CliRunner,
                             mock_setup_session: None) -> None:
    mocker.patch('youtube_unofficial.client.YouTubeClient.get_playlist_video_ids',
                 return_value=[{
                     'z': 1,
                     'fake': 'object'
                 }])
    result = runner.invoke(main, ['print-playlist', '1', '--json'])
    assert result.exit_code == 0
    assert '{"fake": "object", "z": 1}' in result.output


def test_print_history(mocker: MockerFixture, runner: CliRunner, mock_setup_session: None) -> None:
    mocker.patch('youtube_unofficial.client.YouTubeClient.get_history_video_ids',
                 return_value=['1234'])
    result = runner.invoke(main, ['print-history'])
    assert result.exit_code == 0
    assert '1234\n' in result.output


def test_print_history_json(mocker: MockerFixture, runner: CliRunner,
                            mock_setup_session: None) -> None:
    mocker.patch('youtube_unofficial.client.YouTubeClient.get_history_video_ids',
                 return_value=[{
                     'z': 1,
                     'fake': 'object'
                 }])
    result = runner.invoke(main, ['print-history', '--json'])
    assert result.exit_code == 0
    assert '{"fake": "object", "z": 1}' in result.output


def test_remove_history_entries(mocker: MockerFixture, runner: CliRunner,
                                mock_setup_session: None) -> None:
    method = mocker.patch('youtube_unofficial.client.YouTubeClient.remove_video_ids_from_history')
    result = runner.invoke(main, ['remove-history-entries', '1', '2'])
    assert result.exit_code == 0
    assert method.call_count == 2


def test_remove_watch_later_id(mocker: MockerFixture, runner: CliRunner,
                               mock_setup_session: None) -> None:
    method = mocker.patch('youtube_unofficial.client.YouTubeClient.remove_video_id_from_playlist')
    result = runner.invoke(main, ['remove-watch-later-video-id', '1', '2'])
    assert result.exit_code == 0
    assert method.call_count == 2


def test_remove_video_id(mocker: MockerFixture, runner: CliRunner,
                         mock_setup_session: None) -> None:
    method = mocker.patch('youtube_unofficial.client.YouTubeClient.remove_video_id_from_playlist')
    result = runner.invoke(main, ['remove-video-id', 'id', '1', '2'])
    assert result.exit_code == 0
    assert method.call_count == 2


def test_toggle_watch_history(mocker: MockerFixture, runner: CliRunner,
                              mock_setup_session: None) -> None:
    method = mocker.patch('youtube_unofficial.client.YouTubeClient.toggle_watch_history')
    result = runner.invoke(main, ['toggle-watch-history'])
    assert result.exit_code == 0
    assert method.call_count == 1
    assert 'Watch history toggled.' in result.output


def test_toggle_watch_history_error(mocker: MockerFixture, runner: CliRunner,
                                    mock_setup_session: None) -> None:
    method = mocker.patch('youtube_unofficial.client.YouTubeClient.toggle_watch_history',
                          return_value=False)
    result = runner.invoke(main, ['toggle-watch-history'])
    assert result.exit_code == 1
    assert method.call_count == 1
    assert 'Failed to toggle watch history.' in result.output
    assert 'Watch history toggled.' not in result.output


def test_clear_watch_history(mocker: MockerFixture, runner: CliRunner,
                             mock_setup_session: None) -> None:
    method = mocker.patch('youtube_unofficial.client.YouTubeClient.clear_watch_history',
                          return_value=False)
    result = runner.invoke(main, ['clear-watch-history'])
    assert result.exit_code == 0
    assert method.call_count == 1
    assert 'Watch history cleared.' in result.output


def test_clear_watch_later(mocker: MockerFixture, runner: CliRunner,
                           mock_setup_session: None) -> None:
    method = mocker.patch('youtube_unofficial.client.YouTubeClient.clear_watch_later',
                          return_value=False)
    result = runner.invoke(main, ['clear-watch-later'])
    assert result.exit_code == 0
    assert method.call_count == 1
    assert 'Watch later queue cleared.' in result.output



from __future__ import annotations

from typing import TYPE_CHECKING
import json

import pytest

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture
    from requests_mock import Mocker
    from youtube_unofficial.client import YouTubeClient


def test_get_playlist_info(mocker: MockerFixture, requests_mock: Mocker, client: YouTubeClient,
                           data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'VISITOR_DATA': 'test_visitor_data',
                     'USER_SESSION_ID': 'test_session_id',
                     'SESSION_INDEX': 0,
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads((data_path / 'get-playlist-info/00.json').read_text()))
    requests_mock.get('https://www.youtube.com/playlist?list=test_playlist', text='<html></html>')
    result = list(client.get_playlist_info(playlist_id='test_playlist'))
    assert result == [{'playlistVideoRenderer': {}}]


def test_get_playlist_info_continuation(mocker: MockerFixture, requests_mock: Mocker,
                                        client: YouTubeClient, data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'VISITOR_DATA': 'test_visitor_data',
                     'INNERTUBE_CONTEXT_CLIENT_VERSION': '1.0',
                     'USER_SESSION_ID': 'test_session_id',
                     'SESSION_INDEX': 0,
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'get-playlist-info/10-continuation.json').read_text()))
    requests_mock.get('https://www.youtube.com/playlist?list=test_playlist', text='<html></html>')
    requests_mock.post(
        'https://www.youtube.com/test_url?prettyPrint=false',
        response_list=json.loads(
            (data_path / 'get-playlist-info/10-continuation-responses.json').read_text()))
    result = list(client.get_playlist_info(playlist_id='test_playlist'))
    assert result == [{
        'playlistVideoRenderer': {}
    }, {
        'playlistVideoRenderer': {}
    }, {
        'playlistVideoRenderer': {}
    }]


def test_get_playlist_info_empty(mocker: MockerFixture, requests_mock: Mocker,
                                 client: YouTubeClient, data_path: Path) -> None:
    requests_mock.get('https://www.youtube.com/playlist?list=test_playlist', text='<html></html>')
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'VISITOR_DATA': 'test_visitor_data',
                     'USER_SESSION_ID': 'test_session_id',
                     'SESSION_INDEX': 0,
                 })
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'get-playlist-info/10-empty.json').read_text()))
    with pytest.raises(KeyError, match=r'This playlist might be empty\.'):
        list(client.get_playlist_info(playlist_id='test_playlist'))


def test_get_playlist_info_empty_alt(mocker: MockerFixture, requests_mock: Mocker,
                                     client: YouTubeClient) -> None:
    requests_mock.get('https://www.youtube.com/playlist?list=test_playlist', text='<html></html>')
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'VISITOR_DATA': 'test_visitor_data',
                     'USER_SESSION_ID': 'test_session_id',
                     'SESSION_INDEX': 0,
                 })
    mocker.patch('youtube_unofficial.client.initial_data', return_value={'contents': {}})
    with pytest.raises(KeyError, match=r'twoColumnBrowseResultsRenderer'):
        list(client.get_playlist_info(playlist_id='test_playlist'))



"""Typed dictionary for YouTube configuration (ytcfg)."""
from __future__ import annotations

from typing_extensions import TypedDict

__all__ = ('YtcfgDict',)


class YtcfgDict(TypedDict, total=False):
    """Configuration dictionary for YouTube site."""
    DELEGATED_SESSION_ID: str
    """Delegated session ID for a brand account."""
    EVENT_ID: str
    """Unknown."""
    ID_TOKEN: str
    """Unknown."""
    INNERTUBE_API_KEY: str
    """API key."""
    INNERTUBE_CONTEXT_CLIENT_NAME: str
    """Client name."""
    INNERTUBE_CONTEXT_CLIENT_VERSION: str
    """Client version."""
    INNERTUBE_CONTEXT_GL: str
    """Geolocation."""
    INNERTUBE_CONTEXT_HL: str
    """Language code."""
    LOGGED_IN: str
    """Logged in status."""
    SESSION_INDEX: int
    """Session index."""
    PAGE_CL: int
    """Unknown."""
    USER_SESSION_ID: str
    """User session ID."""
    VARIANTS_CHECKSUM: str
    """Checksum for variants."""
    VISITOR_DATA: str
    """Visitor data."""
    XSRF_TOKEN: str
    """Cross-site request forgery token."""



"""Download utility function."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, cast
import logging

from requests import Request, Session
from typing_extensions import overload

if TYPE_CHECKING:
    from collections.abc import Mapping

__all__ = ('download_page',)
log = logging.getLogger(__name__)


@overload
def download_page(sess: Session,
                  url: str,
                  data: Any = None,
                  method: Literal['get', 'post'] = 'get',
                  headers: Mapping[str, str] | None = None,
                  params: Mapping[str, str] | None = None,
                  json: Any = None,
                  *,
                  return_json: Literal[False]) -> str:  # pragma: no cover
    ...


@overload
def download_page(sess: Session,
                  url: str,
                  data: Any = None,
                  method: Literal['get', 'post'] = 'get',
                  headers: Mapping[str, str] | None = None,
                  params: Mapping[str, str] | None = None,
                  json: Any = None,
                  *,
                  return_json: Literal[True]) -> dict[str, Any]:  # pragma: no cover
    ...


def download_page(sess: Session,
                  url: str,
                  data: Any = None,
                  method: Literal['get', 'post'] = 'get',
                  headers: Mapping[str, str] | None = None,
                  params: Mapping[str, str] | None = None,
                  json: Any = None,
                  *,
                  return_json: bool = False) -> str | dict[str, Any]:
    """Download a page using the provided session."""
    if headers:
        sess.headers.update(headers)
    req = Request(method, url, data=data, params=params, json=json)
    prepped = sess.prepare_request(req)
    del prepped.headers['accept-encoding']
    r = sess.send(prepped)
    r.raise_for_status()
    if not return_json:
        return r.text.strip()
    return cast('dict[str, Any]', r.json())



"""Constants."""
from __future__ import annotations

__all__ = ('EXTRACTED_THUMBNAIL_KEYS', 'HISTORY_ENTRY_KEYS_TO_SKIP', 'SIMPLE_TEXT_KEYS',
           'TEXT_RUNS_KEYS', 'THUMBNAILS_KEYS', 'USER_AGENT', 'WATCH_HISTORY_URL',
           'WATCH_LATER_URL')

USER_AGENT = ('Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) '
              'Chrome/136.0.0.0 Safari/537.36')
"""
User agent.

:meta hide-value:
"""
WATCH_HISTORY_URL = 'https://www.youtube.com/feed/history'
"""
URL for the Watch History page.

:meta hide-value:
"""
WATCH_LATER_URL = 'https://www.youtube.com/playlist?list=WL'
"""
URL for the Watch Later playlist.

:meta hide-value:
"""
# print-history-ids constants
EXTRACTED_THUMBNAIL_KEYS = {'width', 'height', 'url'}
HISTORY_ENTRY_KEYS_TO_SKIP = {
    'isWatched', 'menu', 'navigationEndpoint', 'showActionMenu', 'thumbnailOverlays',
    'trackingParams'
}
"""
Keys to skip when extracting history entry data.

:meta hide-value:
"""
SIMPLE_TEXT_KEYS = {
    'shortViewCountText': 'short_view_count_text',
    'viewCountText': 'view_count_text'
}
"""
Keys for simple text fields.

:meta hide-value:
"""
TEXT_RUNS_KEYS = {
    'descriptionSnippet': 'description',
    'longBylineText': 'long_byline_text',
    'ownerText': 'owner_text',
    'title': 'title',
    'shortBylineText': 'short_byline_text',
}
"""
Keys for text runs fields.

:meta hide-value:
"""
THUMBNAILS_KEYS = {'channelThumbnailSupportedRenderers', 'richThumbnail'}
"""
Keys for thumbnail fields.

:meta hide-value:
"""



from __future__ import annotations

from typing import TYPE_CHECKING
import json

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture
    from requests_mock import Mocker
    from youtube_unofficial.client import YouTubeClient


def test_get_playlist_video_ids(mocker: MockerFixture, requests_mock: Mocker, client: YouTubeClient,
                                data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'VISITOR_DATA': 'test_visitor_data',
                     'USER_SESSION_ID': 'test_session_id',
                     'SESSION_INDEX': 0,
                 })
    requests_mock.get('https://www.youtube.com/playlist?list=test_playlist', text='<html></html>')
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'get-playlist-video-ids/20-video-ids.json').read_text()))
    result = list(client.get_playlist_video_ids('test_playlist'))
    assert result == ['test_video_id']


def test_get_playlist_video_ids_dict(mocker: MockerFixture, requests_mock: Mocker,
                                     client: YouTubeClient, data_path: Path) -> None:
    mocker.patch('youtube_unofficial.client.find_ytcfg',
                 return_value={
                     'INNERTUBE_API_KEY': 'test_api_key',
                     'VISITOR_DATA': 'test_visitor_data',
                     'USER_SESSION_ID': 'test_session_id',
                     'SESSION_INDEX': 0,
                 })
    requests_mock.get('https://www.youtube.com/playlist?list=test_playlist', text='<html></html>')
    mocker.patch('youtube_unofficial.client.initial_data',
                 return_value=json.loads(
                     (data_path / 'get-playlist-video-ids/20-video-ids-dict.json').read_text()))
    result = list(client.get_playlist_video_ids('test_playlist', return_dict=True))
    assert result == [{
        'owner': 'test_channel_name',
        'title': 'test_video_title',
        'video_id': 'test_video_id',
        'watch_url': 'https://www.youtube.com/watch?v=test_video_id'
    }, {
        'owner': 'test_channel_name',
        'title': 'test_video_title',
        'video_id': 'test_video_id',
        'watch_url': 'https://www.youtube.com/watch?v=test_video_id'
    }, {
        'owner': None,
        'title': None,
        'video_id': 'test_video_id',
        'watch_url': 'https://www.youtube.com/watch?v=test_video_id'
    }]



"""Utility functions."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar, cast
import json
import logging
import re

from more_itertools import first

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping

    from bs4 import BeautifulSoup as Soup

    from .typing.history import DescriptionSnippet
    from .typing.ytcfg import YtcfgDict

__all__ = ('context_client_body', 'extract_keys', 'find_ytcfg', 'get_text_runs', 'initial_data',
           'ytcfg_headers')

_K = TypeVar('_K')
_V = TypeVar('_V')
log = logging.getLogger(__name__)


def extract_script_content(soup: Soup) -> Iterator[str]:
    yield from (''.join(list(x.children)) for x in soup.select('script'))  # type: ignore[arg-type]


def extract_keys(keys: Iterable[_K], obj: Mapping[_K, _V]) -> dict[_K, _V]:
    """Extract only the specified keys from a dictionary."""
    new = {}
    for key in keys:
        new[key] = obj[key]
    return new


def get_text_runs(desc: DescriptionSnippet) -> str:
    """Extract text from a description snippet."""
    return ''.join(x['text'] for x in desc['runs']).strip().replace('\n', ' - ')


def context_client_body(ytcfg: YtcfgDict) -> dict[str, str]:
    """Get the context client body for API requests."""
    assert 'INNERTUBE_CONTEXT_CLIENT_VERSION' in ytcfg
    return {
        'clientName': 'WEB',
        'clientVersion': ytcfg['INNERTUBE_CONTEXT_CLIENT_VERSION'],
    }


_YT_INITIAL_DATA_RE = r'^var ytInitialData(?:\s+)?='


def initial_data(content: Soup) -> dict[str, Any]:
    """Extract ytInitialData from a BeautifulSoup object."""
    return cast(
        'dict[str, Any]',
        json.loads(
            first(
                re.sub(
                    _YT_INITIAL_DATA_RE, '',
                    first(x for x in extract_script_content(content)
                          if re.match(_YT_INITIAL_DATA_RE, x))).split('\n'))[:-1]))


def find_ytcfg(soup: Soup) -> YtcfgDict:
    """Extract ytcfg from a BeautifulSoup object."""
    return cast(
        'YtcfgDict',
        first(json.JSONDecoder().raw_decode(
            re.sub(r'.+ytcfg.set\(\{',
                   '{',
                   first(x for x in extract_script_content(soup)
                         if '"INNERTUBE_CONTEXT_CLIENT_VERSION":' in x).strip().replace('\n', ''),
                   count=1))))


def ytcfg_headers(ytcfg: YtcfgDict) -> dict[str, str]:
    """Get headers required for API requests."""
    assert 'DELEGATED_SESSION_ID' in ytcfg or 'USER_SESSION_ID' in ytcfg
    return {
        'x-goog-authuser': '0',
        'x-goog-page-id': str(ytcfg.get('DELEGATED_SESSION_ID', ytcfg.get('USER_SESSION_ID', ''))),
        'x-origin': 'https://www.youtube.com'
    }
